/* gencamlog-mov.cc - Insert camera data into a Player logfile for
 *                    playback. Camera images are read from a movie
 *                    file.
 *
 * Scott Livingston
 * Originally created: 25 Feb 2009
 * Current: 4 Mar 2009
 */

#include <cv.h>
#include <highgui.h>

#include <cstdlib> //atoi and atof
#include <iostream>
#include <fstream>
#include <iomanip>
using namespace std;


#define BUF_SIZE 4096


int main( int argc, char **argv )
{
	IplImage *img;
	CvCapture *capt = NULL;
	ifstream imgTimesFile;
	double imgTime;
	int imgNumber;
	double imgBaseTime = -1.;

	/* "current time" in original logfile (i.e., one on which the new
	   logfile is based). This is necessary to ensure proper ordering
	   of original sensor data and inserted camera log entries. */
	double currentOrigLogTime;
	char line[BUF_SIZE]; // line from the logfile
	bool line_printed;

	int camWidth, camHeight, camDepth;

	double logTime;
	double timeOffset;
	ifstream oldLogFile;
	ofstream newLogFile;

	/* Process command-line arguments: (previously existing) logfile
	   and time offset (relative to first time listed in given
	   logfile). */
	if (argc != 6) {
		cout << "Usage: " << argv[0] << " movie-file camdump-times-file time-offset input-logfile output-logfile" << endl;
		return 1;
	}

	if ((capt = cvCreateFileCapture( argv[1] )) == NULL) {
		cerr << "Error: could not open video file for reading: " << argv[1] << endl;
		return -1;
	}
	imgTimesFile.open( argv[2] );

	oldLogFile.open( argv[4] ); // original file on which to base new one
	newLogFile.open( argv[5] ); // file in which to write new logfile

	timeOffset = atof( argv[3] );

	/* NOTE: avoid cvGetCaptureProperty! It tends to return 0 rather
	   than the desired value. */
	//camWidth = (int)cvGetCaptureProperty( capt, CV_CAP_PROP_FRAME_WIDTH );
	//camHeight = (int)cvGetCaptureProperty( capt, CV_CAP_PROP_FRAME_HEIGHT );
	camWidth = 320;
	camHeight = 240;

	/* Initialize currentOrigLogTime (i.e., step to first sensor data
	   entry in the given logfile. */
	while (true) {
		*line = '\0';
		oldLogFile.getline( line, BUF_SIZE );
		if (oldLogFile.eof() || (strlen(line) > 0 && *line != '#')) {
			break;
		} else {
			newLogFile << line << endl;
		}
	}
	if (oldLogFile.eof() && *line == '\0') {
		cerr << "Error: given logfile is empty." << endl;
		return -1;
	}
	currentOrigLogTime = atof(line);
	if (currentOrigLogTime < 0.) {
		cerr << "Error: given logfile indicates negative time." << endl;
		return -1;
	}
	line_printed = false;

	while (cvGetCaptureProperty( capt, CV_CAP_PROP_POS_AVI_RATIO ) < 1) {

		// Extract current image from given movie file
		img = cvQueryFrame( capt );

		// Determine corresponding time from "image times" file
		imgTimesFile >> imgTime >> imgNumber;

		// Stop if end of "image times" file is reached
		if (imgTimesFile.eof())
			break;

		if (imgBaseTime < 0)
			imgBaseTime = imgTime;

		// Insert entry into logfile:

		/* Step through given logfile until we reach appropriate
		   (w.r.t. time) insertion point */
		while (timeOffset+imgTime-imgBaseTime > currentOrigLogTime && !oldLogFile.eof()) {
			newLogFile << line << endl;
			*line = '\0';
			oldLogFile.getline( line, BUF_SIZE );
			if (oldLogFile.eof()) {
				break;
			} else if (*line != '\0') {
				currentOrigLogTime = atof(line);
				if (currentOrigLogTime < 0.) {
					cerr << "Error: given logfile indicates negative time." << endl;
					return -1;
				}
			}
			line_printed = false;
		}

		// common header
		newLogFile << fixed << timeOffset+imgTime-imgBaseTime << ' '
				   << 16777343 << ' ' << 6665 << ' ' // host robot
				   << "camera "
				   << "00 001 001 "; //index type subtype
		
		// camera header
		newLogFile << camWidth << ' ' << camHeight << ' '
				   << 8 << ' '  // image depth (bits per pixel)
				   << 5 << ' '  // 24-bit color (RGB)
				   << 0 << ' '; // no compression

		/*
		// camera data (hex in ASCII!)
		newLogFile.setf( ios::hex );
		for (i = 0; i < camWidth; i++) {
			for (j = 0; j < camHeight; j++) {

				newLogFile << img->imageData[camWidth * j*3 + i*3]
						   << img->imageData[camWidth * j*3 + i*3 + 1]
						   << img->imageData[camWidth * j*3 + i*3 + 2];

			}
		}
		newLogFile.unsetf( ios::hex );
		*/
		
		// corresponding image filename
		newLogFile << "camera_samples/camdat-" << setw(4) << setfill( '0' ) << imgNumber;
		newLogFile << ".jpg" << endl;

	}

	// Copy remainder of given logfile
	while (!oldLogFile.eof()) {
		oldLogFile.getline( line, BUF_SIZE );
		newLogFile << line << endl;
	}

	// Clean up
	cvReleaseCapture( &capt );
	imgTimesFile.close();
	oldLogFile.close();
	newLogFile.close();

	return 0;
}
