/* gencamlog.cc - Insert camera data into a Player logfile for
 *                playback. Note that the image sequence is assumed to
 *                be stored as a collection of JPEG files with a
 *                common (prefix) naming scheme. E.g.,
 *                camdat-0001.jpg, camdat-0002.jpg, ...
 *
 * NOTES: - Assumes image numbering width is 4 digits
 *          (zero-filled). Thus, given image filename prefix PREFIX
 *          and frame number NUM, the filename stored in the logfile
 *          is of the form
 *
 *          PREFIXZNUM.jpg
 *
 *          where ZNUM is NUM zero-filled to achieve width 4 digits
 *          (e.g., 15 is mapped to 0015).
 *
 *
 * Scott Livingston
 * Originally created: 25 Feb 2009
 * Current: 29 Mar 2009
 */

#include <cstdlib> //atoi and atof
#include <cstring>
#include <iostream>
#include <fstream>
#include <iomanip>
using namespace std;


#define BUF_SIZE 4096


int main( int argc, char **argv )
{
	ifstream imgTimesFile;
	double imgTime;
	int imgNumber;
	double imgBaseTime = -1.;

	/* "current time" in original logfile (i.e., one on which the new
	   logfile is based). This is necessary to ensure proper ordering
	   of original sensor data and inserted camera log entries. */
	double currentOrigLogTime;
	char line[BUF_SIZE]; // line from the logfile
	bool line_printed;

	int camWidth, camHeight, camDepth;

	double logTime;
	double timeOffset;
	ifstream oldLogFile;
	ofstream newLogFile;

	/* Process command-line arguments: (previously existing) logfile
	   and time offset (relative to first time listed in given
	   logfile). */
	if (argc != 6) {
		cout << "Usage: " << argv[0] << " filename-prefix camdump-times-file time-offset input-logfile output-logfile" << endl;
		return 1;
	}

	imgTimesFile.open( argv[2] );

	oldLogFile.open( argv[4] ); // original file on which to base new one
	newLogFile.open( argv[5] ); // file in which to write new logfile

	timeOffset = atof( argv[3] );

	/* NOTE: avoid cvGetCaptureProperty! It tends to return 0 rather
	   than the desired value. */
	//camWidth = (int)cvGetCaptureProperty( capt, CV_CAP_PROP_FRAME_WIDTH );
	//camHeight = (int)cvGetCaptureProperty( capt, CV_CAP_PROP_FRAME_HEIGHT );
	camWidth = 320;
	camHeight = 240;

	/* Initialize currentOrigLogTime (i.e., step to first sensor data
	   entry in the given logfile. */
	while (true) {
		*line = '\0';
		oldLogFile.getline( line, BUF_SIZE );
		if (oldLogFile.eof() || (strlen(line) > 0 && *line != '#')) {
			break;
		} else {
			newLogFile << line << endl;
		}
	}
	if (oldLogFile.eof() && *line == '\0') {
		cerr << "Error: given logfile is empty." << endl;
		return -1;
	}
	currentOrigLogTime = atof(line);
	if (currentOrigLogTime < 0.) {
		cerr << "Error: given logfile indicates negative time." << endl;
		return -1;
	}
	line_printed = false;

	while (true) {

		// Determine corresponding time from "image times" file
		imgTimesFile >> imgTime >> imgNumber;

		// Stop if end of "image times" file is reached
		if (imgTimesFile.eof())
			break;

		if (imgBaseTime < 0)
			imgBaseTime = imgTime;

		// Insert entry into logfile:

		/* Step through given logfile until we reach appropriate
		   (w.r.t time) insertion point */
		while (timeOffset+imgTime-imgBaseTime > currentOrigLogTime && !oldLogFile.eof()) {
			newLogFile << line << endl;
			*line = '\0';
			oldLogFile.getline( line, BUF_SIZE );
			if (oldLogFile.eof()) {
				break;
			} else if (*line != '\0') {
				currentOrigLogTime = atof(line);
				if (currentOrigLogTime < 0.) {
					cerr << "Error: given logfile indicates negative time." << endl;
					return -1;
				}
			}
			line_printed = false;
		}

		// common header
		newLogFile << fixed << timeOffset+imgTime-imgBaseTime << ' '
				   << 16777343 << ' ' << 6665 << ' ' // host robot
				   << "camera "
				   << "00 001 001 "; //index type subtype
		
		// camera header
		newLogFile << camWidth << ' ' << camHeight << ' '
				   << 8 << ' '  // image depth (bits per pixel)
				   << 5 << ' '  // 24-bit color (RGB)
				   << 0 << ' '; // no compression

		// corresponding image filename
		newLogFile << argv[1] << setw(4) << setfill( '0' ) << imgNumber;
		newLogFile << ".jpg" << endl;

	}

	// Copy remainder of given logfile
	while (!oldLogFile.eof()) {
		oldLogFile.getline( line, BUF_SIZE );
		newLogFile << line << endl;
	}

	// Clean up
	imgTimesFile.close();
	oldLogFile.close();
	newLogFile.close();

	return 0;
}
