/* KernelObjTracker.cpp
 *
 * Scott Livingston <SCL@utk.edu>
 * April 2009
 *
 */


#include <cmath>

#include "KernelObjTracker.hpp"


KernelObjTracker::KernelObjTracker()
{
	// 0 bins should indicate to the other methods that target model is not initialized.
	numBins = 0; 

	/* ... Note also that targetModel will be an empty vector, so
	   setting numBins to zero is probably not needed. */
}


KernelObjTracker::KernelObjTracker( const IplImage *initImg, CvRect initRegion, int numBins,
									double initBandwidth, bool adaptiveBWScaling )
{
	// Basic parameter checking
	if (initImg == NULL || numBins < 1 || initBandwidth <= 0.) {
		numBins = 0;
		return; // Initialization failed; leave target model empty.
	}

	if (!InitTargetModel( initImg, initRegion, numBins, initBandwidth, adaptiveBWScaling )) {
		// Initialization failed; clear whatever was (partially) created by InitTargetModel method
		numBins = 0;
		targetModel.clear();
		return;
	}
}


bool KernelObjTracker::InitTargetModel( const IplImage *initImg, CvRect initRegion, int numBins,
										double initBandwidth, bool adaptiveBWScaling )
{
	IplImage *hsv, *hue;

	int u, v;
	double u_normal, v_normal;
	double C; // Model normalization factor

	// Basic parameter checking
	if (initImg == NULL || numBins < 1 || initBandwidth <= 0.
		|| initRegion.x < 0 || initRegion.y < 0
		|| initRegion.x+initRegion.width >= initImg->width
		|| initRegion.y+initRegion.height >= initImg->height) {
		return false; // Initialization failed; leave target model as is (prior to calling of this method).
	}

	adaptiveBW = adaptiveBWScaling;
	if (adaptiveBW)
		gammaBW = 0.1; // As suggested in the paper

	this->numBins = numBins;
	bandwidth = initBandwidth;

	// Set scale parameters and target center
	hx = initRegion.width/2.;
	hy = initRegion.height/2.;
	targetCenter.x = (initRegion.x+initRegion.width/2.)/hx;
	targetCenter.y = (initRegion.y+initRegion.height/2.)/hy;

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(initImg), 8, 3 );
	hue = cvCreateImage( cvGetSize(initImg), 8, 1 );
	cvCvtColor( initImg, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	// Generate the target model
	targetModel.clear();
	targetModel.resize( numBins, 0 ); // Initialize target model vector, filled with 0
	C = 0.;
	for (u = initRegion.x; u < initRegion.x+initRegion.width; u++) {
		for (v = initRegion.y; v < initRegion.y+initRegion.height; v++) {

			// Normalized coordinates (in unit circle by construction)
			u_normal = (double)(u)/hx - targetCenter.x;
			v_normal = (double)(v)/hy - targetCenter.y;

			targetModel[Hue2Bin( hue, numBins, 255, u, v )] += KernelProfile( u_normal*u_normal + v_normal*v_normal );
			C += KernelProfile( u_normal*u_normal + v_normal*v_normal );
		}
	}
	for (int i = 0; i < numBins; i++)
		targetModel[i] *= C ? 1/C : 0; // ...avoid divide-by-zero

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );

	return true;
}


double KernelObjTracker::Track( const IplImage *img )
{
	IplImage *hsv, *hue;

	std::vector<double> targetCandidate;
	std::vector<point2d_t> candidatePoints;
	const double epsilon = 1.; // Stopping criterion for kernel shift
	int x0, y0, // Target candidate center
		x1[3], y1[3]; // Updated candidate center (used internally for each iteration)
	double weight; // Calculated for each point in target candidate region

	std::vector<point2d_t>::const_iterator it;
	int binIndex;
	double denom, x_next, y_next; // For kernel-shifting step of tracking algorithm

	double prevBandwidth; // For adaptive bandwidth scaling
	double newBandwidth[3];
	double distance[3];  /* Distance (w.r.t. similarity function) of
							localized target candidate for three
							different bandwidth sizes.

						    If adaptive scaling is disabled, then
						    distance[0] contains the result for
						    original bandwidth. */
	int perturbBW; // counter for trying bandwidth variants
	
	if (img == NULL // Ignore empty calls
		|| targetModel.size() < 1) // ...or if target model is uninitialized.
		return -1.;

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(img), 8, 3 );
	hue = cvCreateImage( cvGetSize(img), 8, 1 );
	cvCvtColor( img, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	prevBandwidth = bandwidth;
	for (perturbBW = 0; perturbBW < (adaptiveBW ? 3 : 1); perturbBW++) {

		switch (perturbBW) {
		case 0:
			bandwidth = newBandwidth[perturbBW] = prevBandwidth;
			break;
		case 1:
			bandwidth = newBandwidth[perturbBW] = 1.05*prevBandwidth;
			break;
		case 2:
			// Prevent bandwidth scaling to near zero values.
			bandwidth = newBandwidth[perturbBW] = (prevBandwidth*hx > 64 && prevBandwidth*hy > 64)? 0.95*prevBandwidth : prevBandwidth;
			break;
		}

		// Initialize search to previously recorded target center
		x1[perturbBW] = x0 = targetCenter.x*hx;
		y1[perturbBW] = y0 = targetCenter.y*hy;

		do {
			x0 = x1[perturbBW];
			y0 = y1[perturbBW];

			if (!CalcTargetCandidate( img, &targetCandidate, &candidatePoints, x0, y0 ))
				return -1.; // Failure during target candidate vector computation

			denom = x_next = y_next = 0.;
			for (it = candidatePoints.begin(); it != candidatePoints.end(); it++) {
				binIndex = Hue2Bin( hue, numBins, 255, (int)(it->x), (int)(it->y) );
				weight = sqrt( targetModel[binIndex] * (targetCandidate[binIndex] ? 1/targetCandidate[binIndex] : 0) );

				x_next += (it->x)*weight;
				y_next += (it->y)*weight;
				denom += weight;
			}
			x1[perturbBW] = (int)(x_next/denom);
			y1[perturbBW] = (int)(y_next/denom);

		} while ((x1[perturbBW]-x0)*(x1[perturbBW]-x0) + (y1[perturbBW]-y0)*(y1[perturbBW]-y0) > epsilon);
	
		/* Calculate the distance between target model and selected
		   target candidate (for this bandwidth variant, if adaptive
		   scaling is enabled). */
		denom = 0.;
		for (int i = 0; i < numBins; i++)
			denom += targetModel[i]*targetCandidate[i];
		distance[perturbBW] = sqrt( 1 - sqrt(denom) );

	}

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );

	/* If adaptive scaling is enabled, then select best (exponentially
	   filtered) new bandwidth and use result */
	if (adaptiveBW) {
		int maxNewBW = 0;
		for (int i = 1; i < 3; i++) {
			if (distance[i] < distance[maxNewBW])
				maxNewBW = i;
		}

		targetCenter.x = x1[maxNewBW]/hx;
		targetCenter.y = y1[maxNewBW]/hy;

		bandwidth = gammaBW*newBandwidth[maxNewBW] + (1-gammaBW)*prevBandwidth;

		return distance[maxNewBW];
	} else {
		targetCenter.x = x1[0]/hx;
		targetCenter.y = y1[0]/hy;
		return distance[0];
	}
}


bool KernelObjTracker::SetBandwidth( double newBandwidth )
{
	if (newBandwidth < 0) {
		return false;
	}

	bandwidth = newBandwidth;
	return true;
}


CvRect KernelObjTracker::GetTargetRegion() const
{
	CvRect targetRegion;

	// If target model is empty, then return empty rectangle.
	if (targetModel.size() < 1)
		return cvRect(0,0,0,0);

	targetRegion.x = (targetCenter.x-bandwidth)*hx;
	targetRegion.y = (targetCenter.y-bandwidth)*hy;
	targetRegion.width = bandwidth*2*hx;
	targetRegion.height = bandwidth*2*hy;

	return targetRegion;
}


bool KernelObjTracker::SetInitCenter( int x0, int y0 )
{
	if (x0 < 0 || y0 < 0)
		return false;

	targetCenter.x = x0/hx;
	targetCenter.y = y0/hy;
	return true;
}


///////////////////////////////////////
// Private Methods
///////////////////////////////////////

int KernelObjTracker::Hue2Bin( const IplImage *hue, int numBins, int maxValue, int x, int y )
{
	int colorValue;
	int i;

	if (hue == NULL || numBins < 1 || maxValue < 0
		|| x < 0 || y < 0
		|| x >= hue->width || y >= hue->height)
		return -1;

	colorValue = (unsigned int)(*(hue->imageData + hue->width * y + x));
	for (i = 0; i < numBins && colorValue > maxValue/numBins*(i+1); i++) {}
	return i < numBins ? i : numBins-1;
}


bool KernelObjTracker::CalcTargetCandidate( const IplImage *img, std::vector<double> *targetCandidate,
											 std::vector<point2d_t> *candidatePoints, int x0, int y0 )
{
	IplImage *hsv, *hue;

	point2d_t tempPoint; // used for populating candidatePoints vector

	int u, v;
	double u_normal, v_normal;
	double Ch; // candidate normalization factor

	int width, height; // Limit region considered according to bandwidth

	// Ignore empty/erroneous calls
	if (img == NULL || x0 < 0 || y0 < 0	|| x0 >= img->width || y0 >= img->height)
		return false;

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(img), 8, 3 );
	hue = cvCreateImage( cvGetSize(img), 8, 1 );
	cvCvtColor( img, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	width = 2*bandwidth*hx;
	height = 2*bandwidth*hy;

	/* Generate the target candidate vector and build list of target
	   candidate points (used later in the tracking algorithm). */
	targetCandidate->clear();
	targetCandidate->resize( numBins, 0 ); // Initialize the target candidate vector, filled with 0
	Ch = 0.;

	candidatePoints->clear();

	for (u = x0-width/2; u < x0+width/2 && u < hue->width-1; u++) {
		if (u < 0)
			continue;
		for (v = y0-height/2; v < y0+height/2 && v < hue->height-1; v++) {
			if (v < 0)
				continue;

			// Normalized coordinates (in unit circle by construction)
			u_normal = (double)(x0-u)/hx;
			v_normal = (double)(y0-v)/hy;

			(*targetCandidate)[Hue2Bin( hue, numBins, 255, u, v )] +=
				KernelProfile( (u_normal*u_normal + v_normal*v_normal)/(bandwidth*bandwidth) );
			Ch += KernelProfile( (u_normal*u_normal + v_normal*v_normal)/(bandwidth*bandwidth) );

			if ((u_normal*u_normal + v_normal*v_normal)/(bandwidth*bandwidth) <= 1) {
				tempPoint.x = u;
				tempPoint.y = v;
				candidatePoints->push_back( tempPoint );
			}
		}
	}
	for (int i = 0; i < numBins; i++)
		(*targetCandidate)[i] *= Ch ? 1/Ch : 0; // ...avoid divide-by-zero

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );
	
	return true;
}
