/* KernelObjTracker.hpp - Class implementing the algorithm proposed in
 *                        the "Kernel-Based Object Tracking" paper by
 *                        D. Comaniciu et al. (IEEE Trans. on Pattern
 *                        Analysis and Machine Intelligence, 2003).
 *
 * Scott Livingston <SCL@utk.edu>
 * April 2009
 *
 * NOTES: - Unless specified otherwise, a function with bool return
 *          type will return true on success, false on failure.
 *
 *        - Any attempts to access points outside of a given image
 *          will be silently bounded. In particular, setting the
 *          bandwidth to be larger than the main diagonal of the image
 *          will cause ALL pixels in the image to be considered; then
 *          iterating several times over a fixed image while using
 *          adaptive bandwidth scaling MIGHT (i.e., I have not
 *          verified this) allow localization of an otherwise lost
 *          tracked object.
 *
 *        - This isn't directly stated in the paper, but for the
 *          implemented kernel profile (Epanechnikov), the bandwidth h
 *          corresponds to the radius of the target (candidate) circle
 *          in the normalized (from the target model) space. Recall
 *          that normalization is achieved by the horizontal and
 *          vertical scaling factors hx and hy, respectively.
 *
 *          From the above observation, a call to the method
 *          KernelObjTracker::GetTargetRegion returns the bounding
 *          rectangle for this ellipse (which is a circle with radius
 *          h in the normalized space) in the most recently processed
 *          image.
 *
 */

#ifndef _KERNELOBJTRACKER_HPP_
#define _KERNELOBJTRACKER_HPP_


#include <cv.h>

#include <vector>
using namespace std;


/* This structure might be useful if I create a vector of target
   candidates in a given image, rather than just finding one. */
/*typedef struct {
	CvRect rect;
	double BhattCoeff; // Similarity measure based on Bhattacharyya coefficient
} TrackRectB;
*/

class KernelObjTracker {

	typedef unsigned int uint;

	typedef struct {
		double x, y;
	} point2d_t;

public:

	KernelObjTracker();
	KernelObjTracker( const IplImage *initImg, CvRect initRegion, int numBins = 16,
					  double initBandwidth = 1., bool adaptiveBWScaling = true );
	~KernelObjTracker() {}

	// (Re)initialize target model
	bool InitTargetModel( const IplImage *initImg, CvRect initRegion, int numBins = 16,
						  double initBandwidth = 1., bool adaptiveBWScaling = true );

	/* Find target in given image, initializing with previous target
	   candidate region. Returns distance measure, which is defined to
	   be sqrt(1 - r) where r is the sample estimate of the
	   Bhattacharyya coefficient; see Equations (6) and (7) in the
	   paper. The algorithm is defined in Section 4.1. */
	double Track( const IplImage *img );

	// Center point at which to initialize target localization algorithm
	bool SetInitCenter( int x0, int y0 );

	int GetCenterX() const { return (int)(targetCenter.x*hx); }
	int GetCenterY() const { return (int)(targetCenter.y*hy); }

	/* Actually returns smallest rectangle that bounds current
	   targetCenter with bandwidth. */
	CvRect GetTargetRegion() const;

	bool SetBandwidth( double newBandwidth );
	double GetBandwidth() const
	{ return bandwidth; }

private:

	// Kernel profile; Epanechnikov, as used in the paper (see Eq. (12)).
	double KernelProfile( double x )
	{ return x <= 1 ? .5*1*(1+2)*(1-x) : 0; }

	// Calculate a target candidate vector.
	bool CalcTargetCandidate( const IplImage *img, std::vector<double> *targetCandidate,
							  std::vector<point2d_t> *candidatePoints, int x0, int y0 );

	/* Returns the index of the bin to which the pixel element of
	   (single channel) hue image belongs.
	   On error, returns -1.
	   Note that if the color at (x,y) is greater than maxValue, the greatest
	   possible bin number is returned, i.e., numBins-1. */
	int Hue2Bin( const IplImage *hue, int numBins, int maxValue, int x, int y );

	vector<double> targetModel; // Kernel-based target model (derived from color histogram)
	double hx, hy; /* Normalization of target model, used in order to
					  scale the known target (during initialization)
					  into a unit circle. */
	int numBins; // Number of bins to use in color (i.e., hue) histogram

	double bandwidth; // Bandwidth of kernel profile for target candidates

	bool adaptiveBW; // Use adaptive scaling of the kernel profile bandwidth?
	double gammaBW; /* For adaptive bandwidth scaling, apply
					   exponential filter with parameter gammaBW. */
	
	point2d_t targetCenter; // Initial center for target localization algorithm

};


#endif
