/* testCHT.cpp - Example using ColorHistTracker class.
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 23 March 2009
 *
 */


#include <cv.h>
#include <highgui.h>

#include <cstdlib>
#include <iostream>
using namespace std;

#include "ColorHistTracker.hpp"


int main( int argc, char **argv )
{
	IplImage *img, *prevImg,
		*dispImage; // Use bounding boxes for visual feedback
	CvCapture *capture;

	if (argc != 6) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height" << endl;
		return 1;
	}

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Open capture device, by default a webcam at /dev/video0
	capture = cvCaptureFromCAM( -1 );
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH, 640 );
    cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT, 480 );
	if (!capture) {
		cerr << "Error: OpenCV capture object failed to open.\nCheck your webcam..." << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	// Attempt to instantiate and initialize a ColorHistTracker object
	ColorHistTracker cht( img, cvRect( atoi(argv[2]), atoi(argv[3]),
									   atoi(argv[4]), atoi(argv[5]) ),
						  16,     //number of histogram bins
						  .85,     //threshold
						  1024 ); //normalization factor
	cvReleaseImage( &img ); // Init target image is no longer needed.

	// Open a window for visual feedback and enter main (tracking) loop
	cvNamedWindow( "track results", 1 );
	//cvNamedWindow( "prep", 1 );
	prevImg = NULL; // Indicate no previous image yet
	dispImage = NULL;
	while (cvWaitKey(10) != 27) {

		// Pull image from capture device
		if (!cvGrabFrame( capture ))
			break;
		img = cvRetrieveFrame( capture );

		// Handle special case of first loop cycle
		if (prevImg == NULL) {
			prevImg = cvCloneImage( img );
			continue;
		}

		// Color histogram-based tracking
		cht.GetMotionRegions( prevImg, img );
		//cht.MaxNCandidateRegions( 7 ); // Area-based pruning
		//cht.CompareRegions( img );
		cht.CompareDerivRegions( img );
		cht.OrderByArea(); //Area-based sorting of "target regions"
		cht.PruneByArea( 1000 );
		cht.OrderByCorrelation(); //Correlation magnitude-based sorting of "target regions"

		// Draw rectangles on image to show found "target regions"
		if (dispImage)
			cvReleaseImage( &dispImage );
		dispImage = cvCloneImage( img );
		if (cht.GetCount() > 0)
			cout << "Correlations:\n";
		for (uint i = 0; i < cht.GetCount() && i < 2; i++) {
			cvRectangle( dispImage,
						 cvPoint(cht[i].rect.x, cht[i].rect.y),
						 cvPoint(cht[i].rect.x+cht[i].rect.width, cht[i].rect.y+cht[i].rect.height),
						 CV_RGB( 255*cht[i].r, 0, 0 ), 5 );
			cout << cht[i].r << ' ';
		}
		if (cht.GetCount() > 0)
			cout << endl;
		cvShowImage( "track results", dispImage );
		
		// For debugging, freeze frame when at least one "target region" is found
		/*if (cht.GetCount() > 0)
			cvWaitKey(0);
		*/

		// Prepare for next loop cycle
		cvCopy( img, prevImg );
	}

	cvDestroyWindow( "track results" );
	//cvDestroyWindow( "prep" );
	cvReleaseCapture( &capture );
	if (prevImg != NULL)
		cvReleaseImage( &prevImg );

	return 0;
}
