/* testCST.cpp - Example using CvCamShiftTracker class.
 *
 * NOTES: - For definition of CvCamShiftTracker, see cvaux.hpp in the
 *          main include directory for opencv (probably
 *          /usr/include/opencv); it basically encapsulates typical
 *          use of the camshift algorithm as exemplified by the
 *          example file camshiftdemo.oc (distributed with the OpenCV
 *          library).
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 6 April 2009
 *
 */


#include <cv.h>
#include <cv.hpp>
#include <cvaux.hpp>
#include <highgui.h>

#include <iostream>
using namespace std;


int main( int argc, char **argv )
{
	IplImage *img,
		*dispImage; // Use bounding boxes for visual feedback
	CvCapture *capture;
	CvRect cst_result;
	int hdims = 16;

	if (argc != 6) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height" << endl;
		return 1;
	}

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Open capture device, by default a webcam at /dev/video0
	capture = cvCaptureFromCAM( -1 );
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH, 320 );
    cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT, 240 );
	if (!capture) {
		cerr << "Error: OpenCV capture object failed to open.\nCheck your webcam..." << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	// Attempt to instantiate and initialize a CvCamShiftTracker object
	CvCamShiftTracker cst;
	if (!cst.set_window( cvRect( atoi(argv[2]), atoi(argv[3]),
								 atoi(argv[4]), atoi(argv[5]) ) )
		|| !cst.set_hist_dims( 1, &hdims )
		|| !cst.update_histogram( img )) {
		cerr << "Error: Could initialize target region." << endl;
		return -1;
	}
	cvReleaseImage( &img ); // Init target image is no longer needed.

	// Open a window for visual feedback and enter main (tracking) loop
	cvNamedWindow( "track results", 1 );
	//cvNamedWindow( "prep", 1 );
	dispImage = NULL;
	while (cvWaitKey(10) != 27) {

		// Pull image from capture device
		if (!cvGrabFrame( capture ))
			break;
		img = cvRetrieveFrame( capture );

		// camshift tracking
		if (!cst.track_object( img )) {
			cerr << "Error: CvCamShiftTracker::track_object method returned error." << endl;
			return -1;
		}
		cst_result = cst.get_window();

		// Draw rectangles on image to show found "target regions"
		if (dispImage)
			cvReleaseImage( &dispImage );
		dispImage = cvCloneImage( img );
		
		cvRectangle( dispImage,
					 cvPoint(cst_result.x, cst_result.y),
					 cvPoint(cst_result.x+cst_result.width, cst_result.y+cst_result.height),
					 CV_RGB( 255, 0, 0 ), 5 );
		cvShowImage( "track results", dispImage );
		
		// For debugging, freeze frame when at least one "target region" is found
		/*if (cht.GetCount() > 0)
			cvWaitKey(0);
		*/

	}

	cvDestroyWindow( "track results" );
	//cvDestroyWindow( "prep" );
	cvReleaseCapture( &capture );

	return 0;
}

